# -*- coding: utf-8 -*-
# This is here for Python 2.7 support.

# ¯\_(ツ)_/¯
"""
A person shrugging their shoulders to indicate a lack of knowledge about a 
particular topic, or a lack of care about the result of a situation.
"""



from os import remove
from os.path import isfile
from sys import path
from os.path import abspath
from sys import argv

path.append(abspath("."))

from meh import Config, Option, UnsupportedTypeError, ExceptionInConfigError

CONFIG_PATH = "data_types.cfg"

config = Config()
config.add(Option("list", [1, 2, 3]))
config.add(Option("tuple", (1, 2, 3)))
config.add(Option("dict", {"foo" : "baz", "test" : 123}))
config.add(Option("bytes", b"test"))
config.add(Option("string", "test"))
config.add(Option("float", 42.0))
config.add(Option("complex", (1+2j)))
config.add(Option("int", 42))
config.add(Option("boolean", False))
config.add(Option("none", None))

try:
	config = config.load(CONFIG_PATH)
except (IOError, ExceptionInConfigError):
	config.dump(CONFIG_PATH)
	config = config.load(CONFIG_PATH)

print(config.list)
print(config.tuple)
print(config.dict)
print(config.bytes)
print(config.string)
print(config.float)
print(config.complex)
print(config.int)
print(config.boolean)
print(config.none)

config.list = [1, 2, 3, 4]
config.tuple = ("1", "2", "3")
config.dict = {"foo" : 123, "test" : 124.0, "foo1" : "bar"}
config.bytes = b"bytes"
config.string = "string"
config.float = 10.0
config.complex = (2+2j)
config.int = 43
config.boolean = True
config.none = None

try:
	config.list = [1, 2, 3, [lambda x: True]]
except UnsupportedTypeError:
	print("Value validatation succeeded for list")

try:
	config.tuple = ("1", "2", (lambda x: True, ))
except UnsupportedTypeError:
	print("Value validatation succeeded for tuple")

try:
	config.dict = {"foo" : lambda x: True, "test" : 124.0, "foo1" : "bar"}
except UnsupportedTypeError:
	print("Value validatation succeeded for dict")

print("File contents:")
print(config._dumps())

if not (len(argv) == 2 and argv[1] == "keep"):
	remove(CONFIG_PATH)
	if isfile(CONFIG_PATH + "c"):
		remove(CONFIG_PATH + "c")
	


from imp import load_source
from os.path import isfile
from sys import version

NoneType = type(None)

class OptionDuplicateError(IndexError):
	def __init__(self, name):
		super(IndexError, self).__init__("'%s' already exists" % name)	


class OptionNotFoundError(IndexError):
	def __init__(self, name):
		super(IndexError, self).__init__("'%s' does not exist" % name)	


class NameMustBeStringError(Exception):
	def __init__(self):
		super(Exception, self).__init__("option names have to be strings")


class ValidationError(Exception):
	def __init__(self, option):
		super(Exception, self).__init__("invalid value for option '%s'" % option)


class UnsupportedTypeError(TypeError):
	def __init__(self):
		super(TypeError, self).__init__("only list, tuple, dict, bytes, "
			"str, float, complex, int, None, and bool are supported (same " 
			"thing applies to list, dict and tuple contents)")	


class ExceptionInConfigError(Exception):
	"""
	Raised if an exception occurs while importing a config file.

	IN: error (hint: error that occured on import)
	"""
	def __init__(self, error):
		self.error = error
		super(Exception, self).__init__("error occured during config import (%s)" % 
			error.__class__.__name__)

def validate_value(value):
	type_value = type(value)
	if type_value in (list, tuple):
		for element in value:
			if not validate_value(element):
				return False
	elif type_value is dict:
		return validate_value(tuple(value.keys())) and validate_value(tuple(value.values()))
	elif type_value in (bytes, str, float, complex, int, bool, NoneType):
		return True
	else:
		return False
	return True

def make_value(value):
	if validate_value(value):
		if type(value) is str:
			value = '"%s"' % value
		elif type(value) in (list, tuple, dict):
			value = str(value)
		return value
	else:
		raise UnsupportedTypeError()


class _EditableConfig:
	"""
	Automatically created proxy class.

	HINTS: 
	_values: All options with their respective values
	_options: All Option instances that were originally added to 
		the Config instance
	_file: Path to the config file
	_validation_failed: Optional function that's called on a validation error
	_debug: Debug mode on/off (obviously)
	"""
	def __init__(self, values, options, file, validation_failed=None, debug=False):
		self._values = values
		self._options = options
		self._file = file
		self._validation_failed = validation_failed
		self._debug = debug


	def __getattr__(self, name):
		if name in self._values:
			return self._values[name]
		else:
			raise AttributeError("config no attribute '%s'" % name)


	def __setattr__(self, name, value):
		if name == "_values" or name not in self._values:
			self.__dict__[name] = value
		else:
			dump_required = False
			for option in self._options:
				if option.name == name:
					if validate_value(value):
						if option.validator != None:
							if option.validator(value):
								self._values[name] = value
								dump_required = True
							else:
								if self._validation_failed != None:
									self._validation_failed(option.name, value)
								else:
									raise ValidationError(option.name)
						else:
							self._values[name] = value
							dump_required = True
					else:
						raise UnsupportedTypeError()
			if dump_required:
				if self._debug: 
					print("Rewriting config because the value of '%s' changed." % name)
				open(self._file, "w").write(self._dumps())


	def _dumps(self):
		out = ""
		for option in self._options:
			value = make_value(self._values[option.name])
			out += "%s = %s%s\n" % (option.name, value, 
				(" # %s" % option.comment) if option.comment else "")
		return out.rstrip("\n")


	def __repr__(self):
		return self._dumps()


class Option:
	
	def __init__(self, name, default_value, validator=None, comment=""):
		if not type(name) is str:
			raise NameMustBeStringError()
		if name.startswith("__"):
			raise InvalidOptionName()
		self._name = name
		self.default_value = default_value
		self.validator = validator
		self.comment = comment


	@property
	def name(self):
		return self._name

	@name.setter
	def name(self, value):
		if name.startswith("__"):
			raise InvalidOptionName()
		self._name = value		


	def __eq__(self, other):
		if other.__class__ == Option:
			return self.__dict__ == other.__dict__


	def __repr__(self):
		return "%s = %s" % (self.name, str(self.default_value)) 


class Config:
	"""
	The central element of Meh (TM).
	IN: 
		options=[] (type: list, hint: a list of options)
		validation_failed=None (type: function, hint: function accepting two 
			parameters that's called when a validation fails)
	
	Example usage:
	from Meh import Config, Option
	config = Config()
	config.add(Option("number", 42, validator=lambda x: type(x) is int))

	CONFIG_PATH = "awesome_config.cfg"
	try:
		config = config.load(CONFIG_PATH)
	except IOError:
		config.dump(CONFIG_PATH)
		config = config.load(CONFIG_PATH)

	print(config.number)
	"""
	def __init__(self, options=[], validation_failed=None, debug=False):
		if type(options) in (list, tuple):
			for option in options:
				if not option.__class__ == Option:
					raise TypeError("all options must be of type Option")
		else:
			raise TypeError("options must be a list or tuple containing options of type Option")
		self.options = options
		self.validation_failed = validation_failed
		self.debug = debug
		self._iterator_index = 0


	def __iter__(self):
		return self


	def __next__(self):
		if self._iterator_index < len(self.options):
			self._iterator_index += 1
			return self.options[self._iterator_index - 1]
		self._iterator_index = 0
		raise StopIteration


	def load(self, file):
		"""
		Returns the actual read- and editable config
		IN: file (type: str, hint: should be a valid path)
		"""
		if isfile(file):
			try:
				config = load_source("config", file)
			except Exception as e:
				raise ExceptionInConfigError(e)
			option_missing = False
			values = {}
			for option in self.options:
				# Make sure all options are avaliable (validators aren't run in this case
				# because there are no values defined)
				if option.name not in dir(config):
					values[option.name] = option.default_value
					option_missing = True
				else:
					# Retrieve the option value
					value = getattr(config, option.name)
					# Make sure validator passes
					if option.validator != None:	
						# If validation doesn't pass
						if not option.validator(value):
							# Resort to default value
							values[option.name] = option.default_value
							if self.validation_failed != None:
								self.validation_failed(option.name, value)
							else:
								raise ValidationError(option.name)
							option_missing = True
						# If validation passes
						else:
							values[option.name] = value
					else:
						values[option.name] = value
				if option_missing:
					self.dump(file)
			return _EditableConfig(values, self.options, file, 
				validation_failed=self.validation_failed, debug=self.debug)
		else:
			error = "'%s' not found" % file
			raise FileNotFoundError(error) if version.startswith("3") else IOError(error)


	def __add__(self, other):
		try:
			self.add(other)
		except TypeError:
			return NotImplemented
		return self


	def __sub__(self, other):
		try:
			self.remove(other)
		except TypeError:
			return NotImplemented
		return self


	def add(self, option):
		"""
		Adds an option to a Config instance
		IN: option (type: Option)
		"""		
		if option.__class__ == Option:
			for _option in self.options:
				if option.name == _option.name:
					raise OptionDuplicateError(_option.name)
			self.options.append(option)
		else:
			raise TypeError("invalid type supplied")


	def remove(self, option):
		"""
		Removes an option from a Config instance
		IN: option (type: Option)
		"""
		if option.__class__ == Option:
			if option in self.options:
				del self.options[self.options.index(option)]
			else:
				raise OptionNotFoundError(option.name)
		else:
			raise TypeError("invalid type supplied")


	def dump(self, file):
		"""
		Writes output of dumps() to the path provided
		IN: file (type: str, hint: should be a valid path)
		"""
		open(file, "w").write(self.dumps())


	def dumps(self):
		"""
		Returns contents of config file as string
		OUT: out (type: str, hint: config content)
		"""
		out = ""
		for option in self.options:
			value = make_value(option.default_value)
			out += "%s = %s%s\n" % (option.name, value, 
				(" # %s" % option.comment) if option.comment else "")
		return out.rstrip("\n")


	def __repr__(self):
		return self.dumps()



import unittest
from tests.validate_collections import ValuesTestCases

def tests():	
	validate_collections_suite = unittest.TestLoader().loadTestsFromTestCase(
		ValuesTestCases)
	return unittest.TestSuite([
		validate_collections_suite,
		])


import unittest
from meh import validate_value


class ValuesTestCases(unittest.TestCase):
	"""Tests for validate_value and make_value"""
	
	def test_can_recursive_lists_be_validated(self):
		"""
		Validate if unsupported data types inside lists in lists detection
		works.
		""" 
		invalid_list = [[lambda x: True, "a", "b", "c", (lambda x: False)], 
				1, 2, lambda x: False]
		valid_list = [1, [1, 2, ("test", )], ["foo"]]
		self.assertFalse(validate_value(invalid_list))
		self.assertTrue(validate_value(valid_list))

	def test_can_lists_be_validated(self):
		"""
		Validate if unsupported data types inside list detection works.
		"""
		invalid_list = [lambda x: False, "a", 1, 2]
		valid_list = [1, 2]
		self.assertFalse(validate_value(invalid_list))
		self.assertTrue(validate_value(valid_list))


	def test_can_recursive_tuples_be_validated(self):
		"""
		Validate if unsupported data types inside lists in lists detection
		works.
		""" 
		invalid_tuple = ((lambda x: True, "a", "b", "c", (lambda x: False)), 
				1, 2, lambda x: False)
		valid_tuple = (1, (1, 2, ("test", )), ("foo"))
		self.assertFalse(validate_value(invalid_tuple))
		self.assertTrue(validate_value(valid_tuple))

	def test_can_tuples_be_validated(self):
		"""
		Validate if unsupported data types inside tuple detection works.
		"""
		invalid_tuple = (lambda x: False, "a", 1, 2)
		valid_tuple = (b"test", 1, 2, 3)
		self.assertFalse(validate_value(invalid_tuple))
		self.assertTrue(validate_value(valid_tuple))

	def test_can_recursive_dicts_be_validated(self):
		"""
		Validate if unsupported data types inside recursive dict key and value 
		combination detection works.
		"""
		invalid_dict = {1 : {"test" : lambda x: True}, 2 : lambda x: True}
		valid_dict = {1 : "foo", 2 : {"baz" : "lol"}}
		self.assertFalse(validate_value(invalid_dict))
		self.assertTrue(validate_value(valid_dict))


	def test_can_dicts_be_validated(self):
		"""
		Validate if unsupported data types inside dict key and value 
		combination detection works.
		"""
		invalid_dict = {1 : "test", 2 : lambda x : True}
		valid_dict = {1 : "foo", 2 : 3}
		self.assertFalse(validate_value(invalid_dict))
		self.assertTrue(validate_value(valid_dict))	


if __name__ == "__main__":
	unittest.main()



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

long_description = """Meh
===

.. figure:: http://content.philip-trauner.me/meh.png
   :alt: Meh Banner


|Python version support: 2.7, 3.4, 3.5| |License: MIT|

***Meh*** is a **Python configuration utility** that uses **Python** as
it’s **data format**. It provides a pleasant and very “pythonic”
interface to do all the things a configuration utility does, like
default values, validators and comments.


.. |Python version support: 2.7, 3.4, 3.5| image:: https://img.shields.io/badge/python-2.7%2C%203.4%2C%203.5-blue.svg
.. |License: MIT| image:: https://img.shields.io/badge/license-MIT-blue.svg
"""

setup(
	name="Meh",
	version="1.2.1",
	author="Philip Trauner",
	author_email="philip.trauner@arztpraxis.io",
	url="https://github.com/PhilipTrauner/Meh",
	packages=find_packages(),
	test_suite="tests.tests.tests",
	description="Python configuration files in Python.",
	long_description=long_description,
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Programming Language :: Python :: 2.7",
		"Programming Language :: Python :: 3.4",
		"Programming Language :: Python :: 3.5",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
	],
	keywords=["config", "cfg", "meh", "py"]
)
